@@ -10,7 +10,7 @@ from apibara.indexer.indexer import Indexer
 from apibara.indexer.info import Info, UserContext
 from apibara.indexer.storage import Filter, IndexerStorage
 from apibara.protocol import StreamService, credentials_with_auth_token
-from apibara.protocol.proto.stream_pb2 import DataFinality
+from apibara.protocol.proto.stream_pb2 import Cursor, DataFinality
 
 logger = logging.getLogger(__name__)
 
@@ -56,6 +56,8 @@ class IndexerRunner(Generic[UserContext, Filter]):
         options to set the input stream and connection string.
     client_options:
         list of options passed to the gRPC channel.
+    timeout:
+        custom timeout for a message to arrive.
     """
 
     def __init__(
@@ -64,6 +66,8 @@ class IndexerRunner(Generic[UserContext, Filter]):
         reset_state: bool = False,
         config: Optional[IndexerRunnerConfiguration] = None,
         client_options: Optional[List[Tuple[str, Any]]] = None,
+        timeout: Optional[int] = None,
+        _reconnect_to_avoid_disconnection: Optional[int] = None,
     ) -> None:
         if config is None:
             config = IndexerRunnerConfiguration()
@@ -73,7 +77,9 @@ class IndexerRunner(Generic[UserContext, Filter]):
         self._config = config
         self._indexer_id = None
         self._indexer_storage = None
+        self._timeout = timeout
         self._client_options = client_options
+        self._reconnect_to_avoid_disconnection = _reconnect_to_avoid_disconnection
 
     async def run(self, indexer: Indexer, *, ctx: Optional[UserContext] = None):
         """Run the indexer until stopped."""
@@ -110,17 +116,27 @@ class IndexerRunner(Generic[UserContext, Filter]):
         while True:
             try:
                 await self._connect_and_stream(indexer, ctx)
+                # stream never ends, except when testing
+                return
             except Exception as exc:
-                logger.debug("indexer exception %A", exc)
+                logger.exception(f"indexer exception")
                 self._retry_count += 1
                 reconnect = await indexer.handle_reconnect(exc, self._retry_count)
 
                 if not reconnect.reconnect:
                     raise
 
-    async def _connect_and_stream(self, indexer: Indexer, ctx: Optional[UserContext]):
+    def _stream_data(self):
+        """Start streaming data.
+
+        This method is provided so that it can be mocked in tests.
+        """
         channel = self._channel()
-        (client, stream) = StreamService(channel).stream_data()
+        (client, stream) = StreamService(channel).stream_data(timeout=self._timeout)
+        return (client, stream, channel)
+
+    async def _connect_and_stream(self, indexer: Indexer, ctx: Optional[UserContext]):
+        (client, stream, _channel) = self._stream_data()
 
         config = indexer.initial_configuration()
         has_stored = self._indexer_storage.update_with_stored_configuration(config)
@@ -141,6 +157,13 @@ class IndexerRunner(Generic[UserContext, Filter]):
 
         previous_end_cursor = None
         pending_received = False
+        additional_filter = None
+        runner_state = "default"  # or "resync"
+
+        # Reconnect every few blocks to avoid disconnection
+        if self._reconnect_to_avoid_disconnection is not None:
+            _blocks_before_reconnect = self._reconnect_to_avoid_disconnection
+
         async for message in stream:
             logger.debug("received message")
             self._retry_count = 0
@@ -150,19 +173,85 @@ class IndexerRunner(Generic[UserContext, Filter]):
                     len(message.data.data) <= 1
                 ), "indexer runner requires batch_size == 1"
 
+                if runner_state == "resync":
+                    logger.debug("handle block resync")
+                    end_cursor = message.data.end_cursor
+                    cursor = message.data.cursor
+                    logger.debug(f"handle resync batch {cursor} - {end_cursor}")
+                    with self._indexer_storage.create_storage_for_data(
+                        message.data.end_cursor
+                    ) as storage:
+                        for batch in message.data.data:
+                            decoded_data = indexer.decode_data(batch)
+                            # only call handler if the batch is for the same block
+                            if (
+                                decoded_data.header.block_number
+                                == previous_end_cursor.order_key
+                            ):
+                                info = Info(
+                                    context=ctx,
+                                    storage=storage,
+                                    cursor=cursor,
+                                    end_cursor=end_cursor,
+                                )
+
+                                await indexer.handle_data(info, decoded_data)
+                                new_additional_filter = indexer._get_and_reset_filter()
+                                if new_additional_filter is not None:
+                                    raise RuntimeError(
+                                        "additional filter not supported when rescanning block"
+                                    )
+                    # in any case, restart syncing from where it left off
+                    runner_state = "default"
+                    config.filter = config.filter.merge(additional_filter)
+                    self._indexer_storage._update_filter(config.filter)
+                    await client.configure(
+                        filter=config.filter.encode(),
+                        finality=config.finality,
+                        cursor=previous_end_cursor,
+                        batch_size=1,
+                    )
+                    continue
+
                 # invalidate any pending data, if any
+                should_invalidate = False
                 if pending_received and previous_end_cursor is not None:
-                    self._indexer_storage.invalidate(previous_end_cursor)
+                    should_invalidate = True
 
                 is_pending = message.data.finality == DataFinality.DATA_STATUS_PENDING
                 pending_received = is_pending
 
                 end_cursor = message.data.end_cursor
                 cursor = message.data.cursor
-                with self._indexer_storage.create_storage_for_data(
-                    message.data.end_cursor
-                ) as storage:
+
+                logger.debug(f"handle batch {cursor} - {end_cursor}")
+
+                if (
+                    self._reconnect_to_avoid_disconnection is not None
+                    and not is_pending
+                ):
+                    _blocks_before_reconnect -= 1
+
+                if is_pending:
+                    create_storage = (
+                        lambda cursor: self._indexer_storage.create_storage_for_pending(
+                            cursor
+                        )
+                    )
+                else:
+                    create_storage = (
+                        lambda cursor: self._indexer_storage.create_storage_for_data(
+                            cursor
+                        )
+                    )
+
+                with create_storage(message.data.end_cursor) as storage:
                     additional_filter = None
+                    if should_invalidate:
+                        self._indexer_storage.invalidate(
+                            previous_end_cursor, session=storage._session
+                        )
+
                     for batch in message.data.data:
                         decoded_data = indexer.decode_data(batch)
                         info = Info(
@@ -173,18 +262,35 @@ class IndexerRunner(Generic[UserContext, Filter]):
                         )
                         if is_pending:
                             await indexer.handle_pending_data(info, decoded_data)
+                            additional_filter = indexer._get_and_reset_filter()
+                            if additional_filter is not None:
+                                raise RuntimeError(
+                                    "additional filter not supported for pending data"
+                                )
                         else:
                             await indexer.handle_data(info, decoded_data)
                             additional_filter = indexer._get_and_reset_filter()
 
-                    if additional_filter is not None:
-                        config.filter = config.filter.merge(additional_filter)
-                        self._indexer_storage._update_filter(
-                            config.filter, session=storage._session
+                        if additional_filter is not None:
+                            logger.debug(
+                                f"filter updated. rescanning block from {end_cursor.order_key - 1}"
+                            )
+                            runner_state = "resync"
+                            await client.configure(
+                                filter=additional_filter.encode(),
+                                finality=config.finality,
+                                cursor=Cursor(order_key=end_cursor.order_key - 1),
+                                batch_size=1,
+                            )
+
+                if not is_pending:
+                    previous_end_cursor = message.data.end_cursor
+
+                if self._reconnect_to_avoid_disconnection is not None:
+                    if _blocks_before_reconnect <= 0:
+                        _blocks_before_reconnect = (
+                            self._reconnect_to_avoid_disconnection
                         )
-                        # user updated the filter.
-                        # TODO: should probably rescan the current block for filters
-                        # that match the new filter.
                         await client.configure(
                             filter=config.filter.encode(),
                             finality=config.finality,
@@ -192,9 +298,6 @@ class IndexerRunner(Generic[UserContext, Filter]):
                             batch_size=1,
                         )
 
-                if not is_pending:
-                    previous_end_cursor = message.data.end_cursor
-
             elif message.invalidate is not None:
                 with self._indexer_storage.create_storage_for_invalidate(
                     message.invalidate.cursor
@@ -203,6 +306,9 @@ class IndexerRunner(Generic[UserContext, Filter]):
                     info = Info(
                         context=ctx, storage=storage, cursor=cursor, end_cursor=cursor
                     )
+
+                    self._indexer_storage.invalidate(cursor, session=storage._session)
+
                     await indexer.handle_invalidate(info, cursor)
                 previous_end_cursor = message.invalidate.cursor
 
