@@ -1,6 +1,6 @@
 import logging
 from contextlib import contextmanager
-from typing import Any, Generic, Iterable, Iterator, Optional, TypeVar
+from typing import Any, Dict, Generic, Iterable, Iterator, Optional, TypeVar
 
 from pymongo import MongoClient
 from pymongo.client_session import ClientSession
@@ -10,10 +10,10 @@ import apibara.cursor as cursor_utils
 from apibara.indexer.configuration import IndexerConfiguration
 from apibara.protocol.proto.stream_pb2 import Cursor
 
-Document = dict[str, Any]
-DocumentFilter = dict[str, Any]
-Update = dict[str, Any]
-Projection = dict[str, any]
+Document = Dict[str, Any]
+DocumentFilter = Dict[str, Any]
+Update = Dict[str, Any]
+Projection = Dict[str, Any]
 
 Filter = TypeVar("Filter")
 
@@ -101,7 +101,7 @@ class IndexerStorage(Generic[Filter]):
 
         return True
 
-    def _update_filter(self, filter: Filter, session: ClientSession):
+    def _update_filter(self, filter: Filter, session: Optional[ClientSession] = None):
         """Set the indexer filter, overriding the previous filter."""
         logger.debug("update stored filter")
 
@@ -111,24 +111,30 @@ class IndexerStorage(Generic[Filter]):
             session=session,
         )
 
-    def invalidate(self, cursor: Cursor):
+    def invalidate(self, cursor: Cursor, session: Optional[ClientSession] = None):
         """Invalidates all data generate after `cursor`."""
-        logger.debug("invalidate data after %A", cursor)
-        with self._mongo.start_session() as session:
-            for collection in self.db.list_collections(session=session):
-                name = collection["name"]
-                if name.startswith("_"):
-                    continue
-                # remove items inserted after block_number
-                self.db[name].delete_many(
-                    {"_chain.valid_from": {"$gt": cursor.order_key}}, session=session
-                )
-                # rollback items updated after block_number
-                self.db[name].update_many(
-                    {"_chain.valid_to": {"$gt": cursor.order_key}},
-                    {"$set": {"_chain.valid_to": None}},
-                    session=session,
-                )
+        logger.debug(f"invalidate data after {cursor}")
+        if session is None:
+            with self._mongo.start_session() as session:
+                self._invalidate(cursor, session)
+        else:
+            self._invalidate(cursor, session)
+
+    def _invalidate(self, cursor: Cursor, session: ClientSession):
+        for collection in self.db.list_collections(session=session):
+            name = collection["name"]
+            if name.startswith("_"):
+                continue
+            # remove items inserted after block_number
+            self.db[name].delete_many(
+                {"_chain.valid_from": {"$gt": cursor.order_key}}, session=session
+            )
+            # rollback items updated after block_number
+            self.db[name].update_many(
+                {"_chain.valid_to": {"$gt": cursor.order_key}},
+                {"$set": {"_chain.valid_to": None}},
+                session=session,
+            )
 
     def drop_database(self):
         logger.debug("dropping database %s", self.db_name)
@@ -163,7 +169,7 @@ class ReadOnlyStorage:
         self,
         collection: str,
         filter: DocumentFilter,
-        sort: Optional[dict[str, int]] = None,
+        sort: Optional[Dict[str, int]] = None,
         projection: Optional[Projection] = None,
         skip: int = 0,
         limit: int = 0,
